@@ -27,7 +27,6 @@
 #include <QQmlContext>
 #include <QQmlEngine>
 #include <QDebug>
-#include <QTimer>
 #include <QPropertyAnimation>
 
 
@@ -38,23 +37,37 @@ class QmlComponentsPoolSingleton
 public:
     QmlComponentsPoolSingleton()
     {}
-
-    QmlComponentsPool self;
+    static QmlComponentsPool *instance(QQmlEngine *engine);
+private:
+    QHash<QQmlEngine*, QmlComponentsPool*> m_instances;
 };
 
 Q_GLOBAL_STATIC(QmlComponentsPoolSingleton, privateQmlComponentsPoolSelf)
 
 
-QmlComponentsPool::QmlComponentsPool(QObject *parent)
-    : QObject(parent)
-{}
-
-void QmlComponentsPool::initialize(QQmlEngine *engine)
+QmlComponentsPool *QmlComponentsPoolSingleton::instance(QQmlEngine *engine)
 {
-    if (!engine || m_instance) {
-        return;
+    Q_ASSERT(engine);
+    auto componentPool = privateQmlComponentsPoolSelf->m_instances.value(engine);
+
+    if (componentPool) {
+        return componentPool;
     }
 
+    componentPool = new QmlComponentsPool(engine);
+
+    QObject::connect(componentPool, &QObject::destroyed, [engine]() {
+        if (privateQmlComponentsPoolSelf) {
+            privateQmlComponentsPoolSelf->m_instances.remove(engine);
+        }
+    });
+    privateQmlComponentsPoolSelf->m_instances[engine] = componentPool;
+    return componentPool;
+}
+
+QmlComponentsPool::QmlComponentsPool(QQmlEngine *engine)
+    : QObject(engine)
+{
     QQmlComponent *component = new QQmlComponent(engine, this);
 
     component->setData(QByteArrayLiteral("import QtQuick 2.7\n"
@@ -274,6 +287,7 @@ ContentItem::ContentItem(ColumnView *parent)
     : QQuickItem(parent),
       m_view(parent)
 {
+    setFlags(flags() | ItemIsFocusScope);
     m_slideAnim = new QPropertyAnimation(this);
     m_slideAnim->setTargetObject(this);
     m_slideAnim->setPropertyName("x");
@@ -322,24 +336,22 @@ void ContentItem::animateX(qreal newX)
 
 void ContentItem::snapToItem()
 {
-    const qreal moveProjection = m_velocity / 1000.0 * 250;
-
     QQuickItem *firstItem = childAt(viewportLeft(), 0);
-
-    m_velocity = 0;
-
     if (!firstItem) {
         return;
     }
-    QQuickItem *nextItem = childItems()[qMin(childItems().length() - 1, childItems().indexOf(firstItem) + 1)];
+    QQuickItem *nextItem = childAt(firstItem->x() + firstItem->width() + 1, 0);
 
     //need to make the last item visible?
-    if (nextItem && width() - (viewportRight()) < viewportLeft() + moveProjection - firstItem->x()) {
+    if (nextItem && ((m_view->dragging() && m_lastDragDelta < 0) ||
+        (!m_view->dragging() && width() - (viewportRight()) < viewportLeft() - firstItem->x()))) {
         m_viewAnchorItem = nextItem;
         animateX(-nextItem->x() + m_leftPinnedSpace);
 
     //The first one found?
-    } else if (viewportLeft() + moveProjection <= firstItem->x() + firstItem->width()/2 || !nextItem) {
+    } else if ((m_view->dragging() && m_lastDragDelta >= 0) ||
+        (!m_view->dragging() && viewportLeft() <= firstItem->x() + firstItem->width()/2) ||
+        !nextItem) {
         m_viewAnchorItem = firstItem;
         animateX(-firstItem->x() + m_leftPinnedSpace);
 
@@ -509,9 +521,15 @@ void ContentItem::updateVisibleItems()
     for (auto *item : m_items) {
         if (item->isVisible() && item->x() + x() < width() && item->x() + item->width() + x() > 0) {
             newItems << item;
+            connect(item, &QObject::destroyed, this, [this, item] {
+                m_visibleItems.removeAll(item);
+            });
         }
     }
 
+    for (auto *item : m_visibleItems) {
+        disconnect(item, &QObject::destroyed, this, nullptr);
+    }
     const QQuickItem *oldFirstVisibleItem = m_visibleItems.isEmpty() ? nullptr : qobject_cast<QQuickItem *>(m_visibleItems.first());
     const QQuickItem *oldLastVisibleItem = m_visibleItems.isEmpty() ? nullptr : qobject_cast<QQuickItem *>(m_visibleItems.last());
 
@@ -552,6 +570,7 @@ void ContentItem::forgetItem(QQuickItem *item)
 
     const int index = m_items.indexOf(item);
     m_items.removeAll(item);
+    disconnect(item, &QObject::destroyed, this, nullptr);
     updateVisibleItems();
     m_shouldAnimate = true;
     m_view->polish();
@@ -568,12 +587,12 @@ QQuickItem *ContentItem::ensureSeparator(QQuickItem *item)
     QQuickItem *separatorItem = m_separators.value(item);
     
     if (!separatorItem) {
-        separatorItem = qobject_cast<QQuickItem *>(privateQmlComponentsPoolSelf->self.m_separatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
+        separatorItem = qobject_cast<QQuickItem *>(privateQmlComponentsPoolSelf->instance(qmlEngine(item))->m_separatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
         if (separatorItem) {
             separatorItem->setParentItem(item);
             separatorItem->setZ(9999);
             separatorItem->setProperty("column", QVariant::fromValue(item));
-            privateQmlComponentsPoolSelf->self.m_separatorComponent->completeCreate();
+            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_separatorComponent->completeCreate();
             m_separators[item] = separatorItem;
         }
     }
@@ -586,12 +605,12 @@ QQuickItem *ContentItem::ensureRightSeparator(QQuickItem *item)
     QQuickItem *separatorItem = m_rightSeparators.value(item);
     
     if (!separatorItem) {
-        separatorItem = qobject_cast<QQuickItem *>(privateQmlComponentsPoolSelf->self.m_rightSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
+        separatorItem = qobject_cast<QQuickItem *>(QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_rightSeparatorComponent->beginCreate(QQmlEngine::contextForObject(item)));
         if (separatorItem) {
             separatorItem->setParentItem(item);
             separatorItem->setZ(9999);
             separatorItem->setProperty("column", QVariant::fromValue(item));
-            privateQmlComponentsPoolSelf->self.m_rightSeparatorComponent->completeCreate();
+            QmlComponentsPoolSingleton::instance(qmlEngine(item))->m_rightSeparatorComponent->completeCreate();
             m_rightSeparators[item] = separatorItem;
         }
     }
@@ -617,7 +636,11 @@ void ContentItem::itemChange(QQuickItem::ItemChange change, const QQuickItem::It
 
         if (!m_items.contains(value.item)) {
             connect(value.item, &QQuickItem::widthChanged, m_view, &ColumnView::polish);
-            m_items << value.item;
+            QQuickItem *item = value.item;
+            m_items << item;
+            connect(item, &QObject::destroyed, this, [this, item]() {
+                m_view->removeItem(item);
+            });
         }
 
         if (m_view->separatorVisible()) {
@@ -706,16 +729,6 @@ ColumnView::ColumnView(QQuickItem *parent)
     setAcceptedMouseButtons(Qt::LeftButton | Qt::BackButton | Qt::ForwardButton);
     setFiltersChildMouseEvents(true);
 
-    m_speedSampler = new QTimer(this);
-    m_speedSampler->setInterval(100);
-    m_speedSampler->setSingleShot(false);
-    connect(m_speedSampler, &QTimer::timeout,
-        this, [this] () {
-            m_contentItem->m_velocity = - (m_contentItem->x() - m_sampledContentX) * 10;
-            m_sampledContentX = m_contentItem->x();
-        }
-    );
-
     connect(m_contentItem->m_slideAnim, &QPropertyAnimation::finished, this, [this] () {
         m_moving = false;
         emit movingChanged();
@@ -723,6 +736,11 @@ ColumnView::ColumnView(QQuickItem *parent)
     connect(m_contentItem, &ContentItem::widthChanged, this, &ColumnView::contentWidthChanged);
     connect(m_contentItem, &ContentItem::xChanged, this, &ColumnView::contentXChanged);
 
+    connect(this, &ColumnView::activeFocusChanged, this, [this]() {
+        if (hasActiveFocus() && m_currentItem) {
+            m_currentItem->forceActiveFocus();
+        }
+    });
     ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(this, true));
     attached->setView(this);
     attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(m_contentItem, true));
@@ -761,7 +779,7 @@ qreal ColumnView::columnWidth() const
 void ColumnView::setColumnWidth(qreal width)
 {
     // Always forget the internal binding when the user sets anything, even the same value
-    disconnect(&privateQmlComponentsPoolSelf->self, &QmlComponentsPool::gridUnitChanged, this, nullptr);
+    disconnect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::gridUnitChanged, this, nullptr);
 
     if (m_contentItem->m_columnWidth == width) {
         return;
@@ -904,7 +922,7 @@ int ColumnView::scrollDuration() const
 
 void ColumnView::setScrollDuration(int duration)
 {
-    disconnect(&privateQmlComponentsPoolSelf->self, &QmlComponentsPool::longDurationChanged, this, nullptr);
+    disconnect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::longDurationChanged, this, nullptr);
 
     if (m_contentItem->m_slideAnim->duration() == duration) {
         return;
@@ -1019,6 +1037,9 @@ void ColumnView::insertItem(int pos, QQuickItem *item)
 
     m_contentItem->m_items.insert(qBound(0, pos, m_contentItem->m_items.length()), item);
 
+    connect(item, &QObject::destroyed, m_contentItem, [this, item]() {
+        removeItem(item);
+    });
     ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
     attached->setOriginalParent(item->parentItem());
     attached->setShouldDeleteOnRemove(item->parentItem() == nullptr && QQmlEngine::objectOwnership(item) == QQmlEngine::JavaScriptOwnership);
@@ -1130,15 +1151,8 @@ QQuickItem *ColumnView::pop(QQuickItem *item)
 
 void ColumnView::clear()
 {
-    ColumnViewAttached *attached = nullptr;
-
     for (QQuickItem *item : m_contentItem->m_items) {
-        attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, false));
-        if (attached && attached->shouldDeleteOnRemove()) {
-            item->deleteLater();
-        } else {
-            item->setParentItem(attached ? attached->originalParent() : nullptr);
-        }
+        removeItem(item);
     }
     m_contentItem->m_items.clear();
     emit contentChildrenChanged();
@@ -1203,11 +1217,6 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
         if (candidateItem->parentItem() == m_contentItem) {
             setCurrentIndex(m_contentItem->m_items.indexOf(candidateItem));
         }
-
-        // Start the speed sampler
-        m_contentItem->m_velocity = 0;
-        m_sampledContentX = m_contentItem->x();
-        m_speedSampler->start();
         break;
     }
     case QEvent::MouseMove: {
@@ -1249,6 +1258,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
             m_contentItem->setBoundedX(m_contentItem->x() + pos.x() - m_oldMouseX);
         }
 
+        m_contentItem->m_lastDragDelta = pos.x() - m_oldMouseX;
         m_oldMouseX = pos.x();
 
         setKeepMouseGrab(m_dragging);
@@ -1276,6 +1286,7 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
         m_mouseDown = false;
 
         m_contentItem->snapToItem();
+        m_contentItem->m_lastDragDelta = 0;
         if (m_dragging) {
             m_dragging = false;
             emit draggingChanged();
@@ -1292,11 +1303,6 @@ bool ColumnView::childMouseEventFilter(QQuickItem *item, QEvent *event)
         setKeepMouseGrab(false);
 
         me->setAccepted(block);
-
-        // Stop the speed sampler
-        m_sampledContentX = m_contentItem->x();
-        m_speedSampler->stop();
-
         return block;
         break;
     }
@@ -1323,9 +1329,6 @@ void ColumnView::mousePressEvent(QMouseEvent *event)
     m_startMouseX = event->localPos().x();
     m_mouseDown = true;
     setKeepMouseGrab(false);
-    m_contentItem->m_velocity = 0;
-    m_sampledContentX = m_contentItem->x();
-    m_speedSampler->start();
     event->accept();
 }
 
@@ -1355,6 +1358,7 @@ void ColumnView::mouseMoveEvent(QMouseEvent *event)
         m_contentItem->setBoundedX(m_contentItem->x() + event->pos().x() - m_oldMouseX);
     }
 
+    m_contentItem->m_lastDragDelta = event->pos().x() - m_oldMouseX;
     m_oldMouseX = event->pos().x();
     event->accept();
 }
@@ -1377,54 +1381,51 @@ void ColumnView::mouseReleaseEvent(QMouseEvent *event)
         return;
     }
 
+    m_contentItem->snapToItem();
+    m_contentItem->m_lastDragDelta = 0;
+
     if (m_dragging) {
         m_dragging = false;
         emit draggingChanged();
     }
 
-    m_sampledContentX = m_contentItem->x();
-    m_speedSampler->stop();
-
-    m_contentItem->snapToItem();
     setKeepMouseGrab(false);
     event->accept();
 }
 
 void ColumnView::mouseUngrabEvent()
 {
-    if (m_dragging) {
-        m_dragging = false;
-        emit draggingChanged();
-    }
-
     m_mouseDown = false;
 
     if (m_contentItem->m_slideAnim->state() != QAbstractAnimation::Running) {
         m_contentItem->snapToItem();
     }
+    m_contentItem->m_lastDragDelta = 0;
+
+    if (m_dragging) {
+        m_dragging = false;
+        emit draggingChanged();
+    }
+
     setKeepMouseGrab(false);
-    m_sampledContentX = m_contentItem->x();
-    m_speedSampler->stop();
 }
 
 void ColumnView::classBegin()
 {
-    privateQmlComponentsPoolSelf->self.initialize(qmlEngine(this));
-
     auto syncColumnWidth = [this]() {
-        m_contentItem->m_columnWidth = privateQmlComponentsPoolSelf->self.m_units->property("gridUnit").toInt() * 20;
+        m_contentItem->m_columnWidth = privateQmlComponentsPoolSelf->instance(qmlEngine(this))->m_units->property("gridUnit").toInt() * 20;
         emit columnWidthChanged();
     };
 
-    connect(&privateQmlComponentsPoolSelf->self, &QmlComponentsPool::gridUnitChanged, this, syncColumnWidth);
+    connect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::gridUnitChanged, this, syncColumnWidth);
     syncColumnWidth();
 
     auto syncDuration = [this]() {
-        m_contentItem->m_slideAnim->setDuration(privateQmlComponentsPoolSelf->self.m_units->property("longDuration").toInt());
+        m_contentItem->m_slideAnim->setDuration(QmlComponentsPoolSingleton::instance(qmlEngine(this))->m_units->property("longDuration").toInt());
         emit scrollDurationChanged();
     };
 
-    connect(&privateQmlComponentsPoolSelf->self, &QmlComponentsPool::longDurationChanged, this, syncDuration);
+    connect(QmlComponentsPoolSingleton::instance(qmlEngine(this)), &QmlComponentsPool::longDurationChanged, this, syncDuration);
     syncDuration();
 
     QQuickItem::classBegin();
@@ -1464,6 +1465,9 @@ void ColumnView::contentChildren_append(QQmlListProperty<QQuickItem> *prop, QQui
     }
 
     view->m_contentItem->m_items.append(item);
+    connect(item, &QObject::destroyed, view->m_contentItem, [view, item]() {
+        view->removeItem(item);
+    });
 
     ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
     attached->setOriginalParent(item->parentItem());
@@ -1531,6 +1535,9 @@ void ColumnView::contentData_append(QQmlListProperty<QObject> *prop, QObject *ob
 
     } else if (item) {
         view->m_contentItem->m_items.append(item);
+        connect(item, &QObject::destroyed, view->m_contentItem, [view, item]() {
+            view->removeItem(item);
+        });
 
         ColumnViewAttached *attached = qobject_cast<ColumnViewAttached *>(qmlAttachedPropertiesObject<ColumnView>(item, true));
         attached->setOriginalParent(item->parentItem());
